#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from enum import Enum

from evo.common.data import OrderByOperatorEnum


def parse_order_by(order_by: dict[Enum | str, OrderByOperatorEnum | str] | None) -> str | None:
    if order_by is None:
        return None
    parts = []
    for field, op in order_by.items():
        if isinstance(field, Enum):
            field = field.value
        if isinstance(op, Enum):
            op = op.value
        parts.append(f"{op}:{field}")
    return ",".join(parts)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from threading import Lock
from typing import BinaryIO

from evo import logging

from .interfaces import IDestination, ISource

logger = logging.getLogger("io.bytes")

__all__ = [
    "BytesDestination",
    "BytesSource",
]


class BytesDestination(IDestination):
    """A wrapper for a writable binary buffer, usually an open file."""

    def __init__(self, backend: BinaryIO):
        self._backend = backend
        self._write_lock = Lock()

    async def write_chunk(self, offset: int, data: bytes) -> None:
        """Write raw data to the destination at the provided offset.

        BytesDestination does not handle any errors from the underlying buffer. Any errors that occur must be handled
        by the calling code.
        """
        logger.debug(f"Writing {len(data)} bytes at offset {offset}")
        with self._write_lock:
            self._backend.seek(offset)
            self._backend.write(data)


class BytesSource(ISource):
    """A wrapper for a readable binary buffer, usually an open file."""

    def __init__(self, backend: BinaryIO, size: int | None = None) -> None:
        """
        :param backend: A readable binary buffer.
        :param size: The size of the source data, or None. If None, the size will be determined by seeking to the end
            of the buffer the first time `get_size()` is called.
        """
        self._backend = backend
        self._size = size
        self._read_lock = Lock()

    async def get_size(self) -> int:
        """Get the size of the source data.

        :returns: The size of the source data.
        """
        if self._size is None:
            with self._read_lock:
                # No need to seek back to the original position, because `read_chunk` will always seek to the correct
                # position before reading.
                self._size = self._backend.seek(0, 2)
        return self._size

    async def read_chunk(self, offset: int, length: int) -> bytes:
        """Read <length> bytes of raw data from the source, starting at the given offset.

        BytesSource does not handle any errors from the underlying buffer. Any errors that occur must be handled
        by the calling code.

        :param offset: The offset in the source data to start reading from.
        :param length: The number of bytes to read.

        :returns: The raw data read from the source.
        """
        logger.debug(f"Reading {length} bytes at offset {offset}")
        with self._read_lock:
            self._backend.seek(offset)
            return self._backend.read(length)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Workspaces API
=============

The Workspaces API enables users to organize, maintain, and store project data, but does not use or process the data. The workspace APIs allow you to manage:
- Workspaces
- User roles within workspaces
- Workspace thumbnails

There are three pre-defined roles within workspaces:

- Owner: can perform all actions in the workspace
- Editor: can perform all actions excluding deleting of a workspace
- Viewer: can view the workspace

These user roles can be assigned to users in a workspace. Once a role has been assigned it can be replaced or removed.
Users can also retrieve user roles, the role of a particular user, and their own role if applicable.
For more information on using the Workspaces API, see the [Workspaces API overview](https://developer.seequent.com/docs/guides/workspaces/), or the API references here.


This code is generated from the OpenAPI specification for Workspaces API.
API version: 1.0
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod
from evo.common.utils import get_header_metadata

from ..models import *  # noqa: F403

__all__ = ["WorkspacesApi"]


class WorkspacesApi:
    """API client for the Workspaces endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def assign_user_role(
        self,
        org_id: str,
        workspace_id: str,
        assign_role_request: AssignRoleRequest,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> UserRole:  # noqa: F405
        """Assign user role

        Assigns a user a role in a workspace.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param assign_role_request:
            Example: `endpoints.AssignRoleRequest()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": UserRole,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/users",
            path_params=_path_params,
            header_params=_header_params,
            body=assign_role_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def create_workspace(
        self,
        org_id: str,
        create_workspace_request: CreateWorkspaceRequest,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> WorkspaceRoleRequiredResponse:  # noqa: F405
        """Create workspace

        Creates a new workspace in your organization.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param create_workspace_request:
            Example: `endpoints.CreateWorkspaceRequest()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": WorkspaceRoleRequiredResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/workspace/orgs/{org_id}/workspaces",
            path_params=_path_params,
            header_params=_header_params,
            body=create_workspace_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def delete_user_role(
        self,
        workspace_id: str,
        user_id: str,
        org_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Remove user from workspace

        Removes a user by removing their role from the workspace.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param user_id:
            Format: `uuid`
            Example: `'user_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "user_id": user_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {} | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.DELETE,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/users/{user_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def delete_workspace(
        self,
        workspace_id: str,
        org_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Delete workspace

        Deletes a workspace. Use of this API requires an \"owner\" role in the workspace. Deleted workspaces will no longer be returned in \"get workspace\" and \"list workspace\" requests, however you can include them by appending `deleted=true` to your request parameters.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {} | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.DELETE,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_current_user_role(
        self,
        workspace_id: str,
        org_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> UserRole:  # noqa: F405
        """Get current user role

        Returns your user account's role in a workspace.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": UserRole,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/current-user-role",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_workspace(
        self,
        org_id: str,
        workspace_id: str,
        deleted: bool | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> WorkspaceRoleRequiredResponse:  # noqa: F405
        """Get workspace

        Returns a workspace by its ID, provided your user has a role to access it. To retrieve a workspace by ID regardless of your user role, see the 'Admin' equivalent API.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param deleted: (optional) Only list workspaces that have been deleted.
            Example: `False`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": WorkspaceRoleRequiredResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_user_roles(
        self,
        workspace_id: str,
        org_id: str,
        filter_user_id: str | None = None,
        user_id: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ListUserRoleResponse:  # noqa: F405
        """List users

        Returns a list of all users with a role in a workspace.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param filter_user_id: (optional) Filter to see the role of a specific user ID.
            Format: `uuid`
            Example: `'filter_user_id_example'`
        :param user_id: (optional) Filter to see the role of a specific user ID.
            Format: `uuid`
            Example: `'user_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if filter_user_id is not None:
            _query_params["filter[user_id]"] = filter_user_id
        if user_id is not None:
            _query_params["user_id"] = user_id

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ListUserRoleResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/users",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_workspaces(
        self,
        org_id: str,
        limit: int | None = None,
        offset: int | None = None,
        sort: str | None = None,
        order_by: str | None = None,
        filter_created_by: str | None = None,
        created_by: str | None = None,
        created_at: str | None = None,
        updated_at: str | None = None,
        filter_name: str | None = None,
        name: str | None = None,
        deleted: bool | None = None,
        filter_user_id: str | None = None,
        user_id: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ListWorkspacesResponse:  # noqa: F405
        """List workspaces

        Returns a paginated list of all workspaces that you have a role to access in your organization. For a list of all workspaces regardless of your user role, see the 'Admin' equivalent API.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param limit: (optional) The maximum number of results to return. Limits over 100 are deprecated and may only work for some clients.
            Example: `20`
        :param offset: (optional) The (zero-based) offset of the first item returned in the collection.
            Example: `0`
        :param sort: (optional) An optional comma separated list of fields to sort the results by. Options are: `name`, `-name`, `created_at`, `-created_at`, `updated_at`, `-updated_at`, `user_role`, `-user_role`.
            Example: `'sort_example'`
        :param order_by: (optional) An optional, comma-separated list of fields by which to order the results. Each field could be prefixed with an order operator: `asc:` for ascending order or `desc:` for descending order, default is ascending order. The sortable fields are: `name`, `created_at`, `updated_at`, and `user_role`. You can also do compound sorts by comma separating your order params, e.g. `asc:user_role,desc:name`. This will sort by user_role in ascending order and then by name in descending order as a secondary sort.
            Example: `'order_by_example'`
        :param filter_created_by: (optional) Filter by workspace that a user has created, by user ID.
            Format: `uuid`
            Example: `'filter_created_by_example'`
        :param created_by: (optional) Filter by workspace that a user has created, by user ID.
            Format: `uuid`
            Example: `'created_by_example'`
        :param created_at: (optional) Filter by the time workspace has created.
            Example: `'created_at_example'`
        :param updated_at: (optional) Filter by the latest time workspace was updated.
            Example: `'updated_at_example'`
        :param filter_name: (optional) Filter by workspace name.
            Example: `'filter_name_example'`
        :param name: (optional) Filter by workspace name.
            Example: `'name_example'`
        :param deleted: (optional) Include workspaces that have been deleted.
            Example: `True`
        :param filter_user_id: (optional) Filter by workspaces that a user ID has access to.
            Format: `uuid`
            Example: `'filter_user_id_example'`
        :param user_id: (optional) Filter by workspaces that a user ID has access to.
            Format: `uuid`
            Example: `'user_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if limit is not None:
            _query_params["limit"] = limit
        if offset is not None:
            _query_params["offset"] = offset
        if sort is not None:
            _query_params["sort"] = sort
        if order_by is not None:
            _query_params["order_by"] = order_by
        if filter_created_by is not None:
            _query_params["filter[created_by]"] = filter_created_by
        if created_by is not None:
            _query_params["created_by"] = created_by
        if created_at is not None:
            _query_params["created_at"] = created_at
        if updated_at is not None:
            _query_params["updated_at"] = updated_at
        if filter_name is not None:
            _query_params["filter[name]"] = filter_name
        if name is not None:
            _query_params["name"] = name
        if deleted is not None:
            _query_params["deleted"] = deleted
        if filter_user_id is not None:
            _query_params["filter[user_id]"] = filter_user_id
        if user_id is not None:
            _query_params["user_id"] = user_id

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ListWorkspacesResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_workspaces_summary(
        self,
        org_id: str,
        limit: int | None = None,
        offset: int | None = None,
        sort: str | None = None,
        order_by: str | None = None,
        filter_created_by: str | None = None,
        created_by: str | None = None,
        created_at: str | None = None,
        updated_at: str | None = None,
        filter_name: str | None = None,
        name: str | None = None,
        deleted: bool | None = None,
        filter_user_id: str | None = None,
        user_id: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ListWorkspaceSummaryResponse:  # noqa: F405
        """List Workspaces Summary


        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param limit: (optional) The maximum number of results to return.
            Example: `56`
        :param offset: (optional) The (zero-based) offset of the first item returned in the collection.
            Example: `0`
        :param sort: (optional) An optional comma separated list of fields to sort the results by. Options are: `name`, `-name`, `created_at`, `-created_at`, `updated_at`, `-updated_at`, `user_role`, `-user_role`.
            Example: `'sort_example'`
        :param order_by: (optional) An optional, comma-separated list of fields by which to order the results. Each field could be prefixed with an order operator: `asc:` for ascending order or `desc:` for descending order, default is ascending order. The sortable fields are: `name`, `created_at`, `updated_at`, and `user_role`. You can also do compound sorts by comma separating your order params, e.g. `asc:user_role,desc:name`. This will sort by user_role in ascending order and then by name in descending order as a secondary sort.
            Example: `'order_by_example'`
        :param filter_created_by: (optional) Filter by workspace that a user has created, by user ID.
            Format: `uuid`
            Example: `'filter_created_by_example'`
        :param created_by: (optional) Filter by workspace that a user has created, by user ID.
            Format: `uuid`
            Example: `'created_by_example'`
        :param created_at: (optional) Filter by the time workspace has created.
            Example: `'created_at_example'`
        :param updated_at: (optional) Filter by the latest time workspace was updated.
            Example: `'updated_at_example'`
        :param filter_name: (optional) Filter by workspace name.
            Example: `'filter_name_example'`
        :param name: (optional) Filter by workspace name.
            Example: `'name_example'`
        :param deleted: (optional) Include workspaces that have been deleted.
            Example: `True`
        :param filter_user_id: (optional) Filter by workspaces that a user ID has access to.
            Format: `uuid`
            Example: `'filter_user_id_example'`
        :param user_id: (optional) Filter by workspaces that a user ID has access to.
            Format: `uuid`
            Example: `'user_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if limit is not None:
            _query_params["limit"] = limit
        if offset is not None:
            _query_params["offset"] = offset
        if sort is not None:
            _query_params["sort"] = sort
        if order_by is not None:
            _query_params["order_by"] = order_by
        if filter_created_by is not None:
            _query_params["filter[created_by]"] = filter_created_by
        if created_by is not None:
            _query_params["created_by"] = created_by
        if created_at is not None:
            _query_params["created_at"] = created_at
        if updated_at is not None:
            _query_params["updated_at"] = updated_at
        if filter_name is not None:
            _query_params["filter[name]"] = filter_name
        if name is not None:
            _query_params["name"] = name
        if deleted is not None:
            _query_params["deleted"] = deleted
        if filter_user_id is not None:
            _query_params["filter[user_id]"] = filter_user_id
        if user_id is not None:
            _query_params["user_id"] = user_id

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ListWorkspaceSummaryResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces/summary",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def restore_soft_deleted_workspace(
        self,
        workspace_id: str,
        org_id: str,
        deleted: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Restore a deleted workspace

        If a workspace has been soft deleted, this API allows you to restore it again. After restoring, the workspace will no longer be included in \"get workspace\" and \"list workspace\" when the `deleted=true` parameter is appended to your request parameters.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param deleted: (optional)
            Example: `'deleted_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
            "303": ErrorResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def update_workspace(
        self,
        org_id: str,
        workspace_id: str,
        update_workspace_request: UpdateWorkspaceRequest,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> WorkspaceRoleRequiredResponse:  # noqa: F405
        """Update workspace

        Updates a workspace.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param update_workspace_request:
            Example: `endpoints.UpdateWorkspaceRequest()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": WorkspaceRoleRequiredResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.PATCH,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}",
            path_params=_path_params,
            header_params=_header_params,
            body=update_workspace_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2026 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Common geometry types shared across Evo SDK packages.

These types provide a lightweight, dependency-free representation of common
3D geometry primitives used by block models, grids, and other spatial objects.
"""

from __future__ import annotations

from typing import NamedTuple

__all__ = [
    "BoundingBox",
    "Point3",
    "Size3d",
    "Size3i",
]


class Point3(NamedTuple):
    """A 3D point defined by X, Y, and Z coordinates."""

    x: float
    y: float
    z: float


class Size3d(NamedTuple):
    """A 3D size defined by dx, dy, and dz dimensions."""

    dx: float
    dy: float
    dz: float


class Size3i(NamedTuple):
    """A 3D size defined by nx, ny, and nz integer dimensions."""

    nx: int
    ny: int
    nz: int

    @property
    def total_size(self) -> int:
        """The total size (number of elements) represented by this Size3i."""
        return self.nx * self.ny * self.nz


class BoundingBox(NamedTuple):
    """An axis-aligned bounding box defined by minimum and maximum coordinates."""

    x_min: float
    x_max: float
    y_min: float
    y_max: float
    z_min: float
    z_max: float

    @classmethod
    def from_origin_and_size(cls, origin: Point3, size: Size3i, cell_size: Size3d) -> BoundingBox:
        """Create a bounding box from an origin point and grid dimensions.

        :param origin: The origin point of the grid.
        :param size: The number of cells in each dimension.
        :param cell_size: The size of each cell in each dimension.
        :return: A BoundingBox enclosing the grid.
        """
        return cls(
            x_min=origin.x,
            x_max=origin.x + size.nx * cell_size.dx,
            y_min=origin.y,
            y_max=origin.y + size.ny * cell_size.dy,
            z_min=origin.z,
            z_max=origin.z + size.nz * cell_size.dz,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from importlib import metadata
from typing import NamedTuple

__all__ = ["get_header_metadata"]


class PackageDetails(NamedTuple):
    name: str
    version: str


@functools.cache
def get_header_metadata(candidate: str) -> dict[str, str]:
    """Get package name and version given the module __name__ for use in headers.

    :param candidate: The module __name__ to start searching from.

    :return: A dictionary with a value containing the package name and version.
    """
    package_details = get_package_details(candidate)
    return {"Evo-SDK-Source": f"{package_details.name}@v{package_details.version}"}


@functools.cache
def get_package_details(candidate: str) -> PackageDetails:
    """Get package name and version given the module __name__.

    Falls back to `evo-sdk-common` if no package is found for the provided module name.

    :param candidate: The module __name__ to start searching from.

    :return: A dictionary containing the package name and version.
    """
    while candidate:
        try:
            package_metadata = metadata.metadata(candidate)
        except metadata.PackageNotFoundError:
            candidate, *_ = candidate.rpartition(".")
        else:
            return PackageDetails(
                name=package_metadata.get("name", "evo-sdk"),
                version=package_metadata.get("version", "0.0.0"),
            )

    return get_package_details("evo-sdk-common")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import base64
import os
from collections.abc import Awaitable, Callable

from evo import logging
from evo.common.interfaces import ITransport

from .._types import PathLike, resolve_path
from ..data import EmptyResponse, RequestMethod
from ..interfaces import IFeedback
from ..utils import Retry
from .bytes import BytesSource
from .chunked_io_manager import ChunkedIOManager
from .exceptions import ChunkedIOError
from .http import HTTPIOBase
from .interfaces import IDestination

logger = logging.getLogger("io.storage")

__all__ = [
    "StorageDestination",
]


class StorageBlock:
    """Metadata about a block in storage."""

    def __init__(self, byte_offset: int) -> None:
        """Create a new block with the given byte offset."""
        self._offset = byte_offset

    @property
    def byte_offset(self) -> int:
        """The byte offset of the block."""
        return self._offset

    @property
    def id(self) -> str:
        """The block id of the block."""
        block_index = f"{self._offset:032d}"
        block_id_bytes = base64.b64encode(block_index.encode("utf-8"))
        return block_id_bytes.decode("utf-8")

    def __str__(self) -> str:
        return f"<Latest>{self.id}</Latest>"


class BlockList:
    """A list of blocks to be committed to storage."""

    def __init__(self) -> None:
        self._mutex = asyncio.Lock()
        self._sealed = False
        self._blocks: list[StorageBlock] = []

    async def add_block(self, byte_offset: int) -> str:
        """Add a block with the given byte offset to the list.

        :param byte_offset: The byte offset of the block.

        :return: The block id of the created block.
        """
        new_block = StorageBlock(byte_offset)
        async with self._mutex:
            assert not self._sealed, "Cannot add block to sealed block list."
            existing_blocks = {block.byte_offset: block for block in self._blocks}
            if byte_offset in existing_blocks.keys():
                # Remove the old block for this offset, duplicates are not allowed.
                self._blocks.remove(existing_blocks[byte_offset])
            self._blocks.append(new_block)
        return new_block.id

    async def prepare(self) -> bytes:
        """Prepare the block list for committing to storage.

        :return: The XML representation of the block list as bytes.
        """
        async with self._mutex:
            self._sealed = True
            return "\n".join(
                [
                    '<?xml version="1.0" encoding="utf-8"?>',
                    "<BlockList>",
                    *(f"  {block}" for block in sorted(self._blocks, key=lambda b: b.byte_offset)),
                    "</BlockList>",
                ]
            ).encode("utf-8")


class StorageDestination(HTTPIOBase, IDestination):
    """`evo.common.io.interfaces.IDestination` implementation for uploading to storage."""

    def __init__(
        self,
        url_callback: Callable[[], Awaitable[str]],
        transport: ITransport,
        additional_headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(url_callback, transport, additional_headers)
        self._block_list = BlockList()
        self._committed = False

    async def write_chunk(self, offset: int, data: bytes) -> None:
        """Write a chunk of data to storage.

        :param offset: Byte number to start writing from.
        :param data: Bytes to write.

        :raises AssertionError: if the data is already committed
        """
        logger.debug(f"Writing {len(data)} at offset {offset}")
        if self._committed:
            raise AssertionError("Cannot write chunk because data is already committed.")
        block_id = await self._block_list.add_block(offset)
        logger.debug(f"Staging block at offset {offset:032d} with id {block_id}")
        await self._query_resource(
            RequestMethod.PUT,
            query_params={"comp": "block", "blockid": block_id},
            header_params={"Content-Length": str(len(data))},
            body=data,
            response_types_map={"201": EmptyResponse},
        )
        logger.debug(f"Staging {block_id} complete")

    async def _commit(self) -> None:
        payload = await self._block_list.prepare()
        logger.debug(f"Committing blocks:\n{payload.decode('utf-8')}")
        await self._query_resource(
            RequestMethod.PUT,
            query_params={"comp": "blocklist"},
            header_params={
                "Content-Type": "text/plain; charset=UTF-8",
                "Content-Length": str(len(payload)),
            },
            body=payload,
            response_types_map={"201": EmptyResponse},
        )

    async def commit(self, retry: Retry | None = None) -> None:
        """Commit all staged blocks to storage.

        :param retry: A Retry object with a wait strategy. If a chunk is successfully transferred the attempt counter
            will be reset.

        :raises ChunkedIOException: if a non-recoverable exception occurred. ChunkedIOException should be caught by
            client code and raised as either ProcessingError or ApplicationError for a user-friendly presentation.
        :raises RetryError: if the maximum number of consecutive attempts have failed. RetryError is a
            subclass of ChunkedIOException.
        """
        logger.debug("Committing data...")
        assert not self._committed, "Cannot commit data because data is already committed."

        if retry is None:
            retry = Retry(logger=logger)

        async for handler in retry:
            with handler.suppress_errors(ChunkedIOError):
                await self._commit()

            if isinstance(handler.exception, ChunkedIOError):
                if not await handler.exception.recover():
                    raise handler.exception

        self._committed = True
        logger.debug("The data was successfully committed.")

    @staticmethod
    async def upload_file(
        filename: PathLike,
        url_generator: Callable[[], Awaitable[str]],
        transport: ITransport,
        max_workers: int | None = None,
        retry: Retry | None = None,
        additional_headers: dict[str, str] | None = None,
        fb: IFeedback | None = None,
    ) -> None:
        """Upload a file with the given filename to the given Storage URL.

        The url generator MUST generate a url that contains a valid SAS (shared access signature) token. The url
        generator may be called again if the last url expires (unless Retry is initialised with max_attempts == 0).

        :param filename: file to upload
        :param url_generator: An awaitable callback that accepts no arguments and returns a URL to download from
        :param transport: The transport to use for the API calls.
        :param max_workers: The maximum number of concurrent connections to use for data transfer. If None, the default
            is `min(32, os.cpu_count() + 4)`
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created. If a chunk is successfully
            transferred the attempt counter will be reset.
        :param additional_headers: Additional headers to include in each request.
        :param fb: feedback to track the upload, by tracking reads from the file only

        :raises ValueError: if the file to upload does not exist
        :raises ChunkedIOException: if a non-recoverable exception occurred.
        :raises RetryError: if the maximum number of consecutive attempts have failed. RetryError is a subclass
            of ChunkedIOException.
        """
        src_path = resolve_path(filename)

        if not src_path.is_file():
            raise ValueError(f"file {src_path} does not exist")

        if retry is None:
            retry = Retry(logger=logger)

        manager = ChunkedIOManager(message="Uploading", retry=retry, max_workers=max_workers)
        async with StorageDestination(
            url_callback=url_generator, transport=transport, additional_headers=additional_headers
        ) as destination:
            size = os.path.getsize(src_path)
            with src_path.open("rb") as input_:
                source = BytesSource(input_, size)
                await manager.run(source, destination, fb)
            await destination.commit(retry=retry)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

try:
    import aiohttp
    from aiohttp.client_exceptions import ClientError
    from aiohttp.typedefs import StrOrURL
except ImportError:
    raise ImportError("AioTransport cannot be used because the aiohttp library is not installed.")
import asyncio
import json
import re
from types import TracebackType

from evo.common import HTTPHeaderDict, HTTPResponse, RequestMethod
from evo.common.exceptions import ClientValueError, TransportError
from evo.common.interfaces import ITransport
from evo.common.utils import BackoffIncremental, BackoffMethod, Retry
from evo.logging import getLogger

from ._helpers import Config, Context

__all__ = ["AioTransport"]

logger = getLogger("aio.transport")

_RE_JSON = re.compile(r"json", re.IGNORECASE)


class AioTransport(ITransport):
    """A client for managing concurrent connections to multiple APIs with a common configuration.

    See `evo.common.interfaces.ITransport` for more detail.
    """

    def __init__(
        self,
        user_agent: str,
        max_attempts: int = 3,
        backoff_method: BackoffMethod = BackoffIncremental(2),
        num_pools: int = 4,
        verify_ssl: bool = True,
        proxy: StrOrURL | None = None,
        close_grace_period_ms: int = 250,
    ):
        """
        :param user_agent: The value to provide in the `User-Agent` header.
        :param max_attempts: Configure the number of retries to allow before raising a `MaxRetryError`
            exception.
        :param backoff_method: Configure the retry backoff implementation.
        :param num_pools: Number of connection pools to cache before discarding the least recently used pool.
        :param verify_ssl: Verify SSL certificates. This should not be disabled unless you really absolutely have to,
            and never in production environments.
        :param close_grace_period_ms: Grace period (in milliseconds) to wait for connections to close.
        """
        self.__config = Config(
            user_agent=user_agent,
            num_pools=num_pools,
            verify_ssl=verify_ssl,
            retry=Retry(logger=logger, max_attempts=max_attempts, backoff_method=backoff_method),
            proxy=proxy,
            close_grace_period_ms=close_grace_period_ms,
        )
        self.__context: Context | None = None
        self.__mutex = asyncio.Lock()
        self.__num_handles = 0

    def __n_handles(self) -> str:
        """Helper method that returns a string describing the number of open handles."""
        assert self.__mutex.locked(), "This method should only be called while holding the mutex."
        return f"{self.__num_handles} handle{'' if self.__num_handles in {1, -1} else 's'}"

    async def open(self) -> None:
        logger.debug("Opening transport.")
        async with self.__mutex:
            if self.__context is None:
                logger.debug("Creating new context.")
                self.__context = await self.__config.create_context()
            self.__num_handles += 1
            logger.debug(f"Transport has {self.__n_handles()}.")
        logger.debug("Transport opened.")

    async def close(self) -> None:
        logger.debug("Closing transport.")
        async with self.__mutex:
            self.__num_handles -= 1
            logger.debug(f"{self.__n_handles()} remaining.")
            if self.__num_handles <= 0:
                logger.debug("Closing context.")
                context = self.__context
                self.__context = None
                await context.close()
            assert self.__num_handles >= 0, "Number of close calls is greater than number of open calls."
        logger.debug("Transport closed.")

    async def __aenter__(self) -> AioTransport:
        await self.open()
        return self

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_val: Exception | None, exc_tb: TracebackType | None
    ) -> None:
        await self.close()

    async def __unwrap(self) -> Context:
        logger.debug("Retrieving context.")
        async with self.__mutex:
            if self.__context is None:
                raise TransportError(
                    "Cannot make a request before the transport has been opened, or after it has been closed."
                )
            else:
                logger.debug("Context retrieved.")
                return self.__context

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> HTTPResponse:
        if post_params is not None and body is not None:
            raise ClientValueError(msg="HTTP body and post parameters cannot be used at the same time.")

        headers = HTTPHeaderDict(headers or {})
        post_params = post_params or []

        if "User-Agent" not in headers:
            headers["User-Agent"] = self.__config.user_agent

        # Resolve timeout value.
        match request_timeout:
            case int() | float():
                timeout = aiohttp.ClientTimeout(total=request_timeout)
            case (sock_connect, sock_read):
                timeout = aiohttp.ClientTimeout(sock_connect=sock_connect, sock_read=sock_read)
            case _:
                timeout = None

        ctx = await self.__unwrap()
        str_method = str(method)
        try:
            match headers.get("Content-Type"), method, body:
                case "application/x-www-form-urlencoded" | "multipart/form-data", _, _:
                    return await ctx.request_form(
                        str_method,
                        url,
                        headers=headers,
                        fields=post_params,
                        timeout=timeout,
                    )
                case _, _, str() | bytes():
                    # Allow any content-type if the body is already serialized.
                    return await ctx.request(str_method, url, headers=headers, body=body, timeout=timeout)
                case content_type, _, _ if content_type is None or _RE_JSON.search(content_type):
                    # Use JSON as the default body encoding.
                    return await ctx.request(
                        str_method,
                        url,
                        headers=headers,
                        body=json.dumps(body) if body else None,
                        timeout=timeout,
                    )
                case _, _, _:
                    raise TransportError(
                        msg="Cannot prepare a request message with the provided arguments. "
                        "Please check that your arguments match the declared content type."
                    )
        except (ClientError, TimeoutError) as e:
            raise TransportError(msg="Could not complete HTTP request", caused_by=e)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from datetime import timezone
from uuid import UUID

from evo.common import ServiceUser

from .data import (
    AddedInstanceUsers,
    BasicWorkspace,
    BoundingBox,
    Coordinate,
    InstanceRole,
    InstanceRoleWithPermissions,
    InstanceUser,
    InstanceUserInvitation,
    InstanceUserWithEmail,
    User,
    UserRole,
    Workspace,
    WorkspaceRole,
)
from .endpoints.models import (
    AddInstanceUsersResponse,
    BaseInstanceUserResponse,
    BasicWorkspaceResponse,
    ListInstanceRolesResponse,
    ListInstanceUserInvitationsResponse,
    WorkspaceRoleOptionalResponse,
    WorkspaceRoleRequiredResponse,
)
from .endpoints.models import BoundingBox as PydanticBoundingBox
from .endpoints.models import Coordinate as PydanticCoordinate
from .endpoints.models import User as PydanticUser
from .endpoints.models import UserRole as PydanticUserRole

__all__ = [
    "bounding_box",
    "instance_user_invitation_model",
    "instance_user_model",
    "instance_user_role_model",
    "instance_user_with_email_model",
    "user_model",
    "user_role_model",
    "workspace_basic_model",
    "workspace_model",
]


def bounding_box(model: PydanticBoundingBox) -> BoundingBox:
    """
    Parse a BoundingBox from the generated model.

    :param model: The model returned by the generated code.
    :return: A BoundingBox instance.
    """

    def convert_coordinate(pydantic_coordinate: PydanticCoordinate) -> Coordinate:
        return Coordinate(latitude=pydantic_coordinate.root[1], longitude=pydantic_coordinate.root[0])

    coordinates = [[convert_coordinate(coord) for coord in coord_list] for coord_list in model.coordinates]

    return BoundingBox(coordinates=coordinates, type=str(model.type.value))


def workspace_model(
    model: WorkspaceRoleOptionalResponse | WorkspaceRoleRequiredResponse, org_id: UUID, base_url: str
) -> Workspace:
    """
    Parse a Workspace from the generated model.
    :param model: The model returned by the generated code.
    :param org_id: The organization ID.
    :param base_url: The base URL of the hub.
    :return: A Workspace instance.
    """
    parsed_bounding_box = None
    if model.bounding_box:
        parsed_bounding_box = bounding_box(model.bounding_box)

    return Workspace(
        id=model.id,
        org_id=org_id,
        hub_url=base_url,
        display_name=model.name,
        description=model.description,
        user_role=WorkspaceRole[str(model.current_user_role.value)] if model.current_user_role else None,
        created_at=model.created_at.replace(tzinfo=timezone.utc),
        created_by=ServiceUser.from_model(model.created_by),
        updated_at=model.updated_at.replace(tzinfo=timezone.utc),
        updated_by=ServiceUser.from_model(model.updated_by),
        bounding_box=parsed_bounding_box,
        default_coordinate_system=model.default_coordinate_system,
        labels=model.labels,
    )


def workspace_basic_model(model: BasicWorkspaceResponse) -> BasicWorkspace:
    """
    Parse a BasicWorkspace from the generated model.
    :param model: The model returned by the generated code.
    :return: A BasicWorkspace instance.
    """
    return BasicWorkspace(
        id=model.id,
        display_name=model.name,
    )


def user_role_model(model: PydanticUserRole) -> UserRole:
    """
    Parse a UserRole from the generated model.
    :param model: The model returned by the generated code.
    :return: A UserRole instance.
    """
    return UserRole(user_id=model.user_id, role=WorkspaceRole[str(model.role.value)])


def user_model(model: PydanticUser) -> User:
    """
    Parse a User from the generated model.
    :param model: The model returned by the generated code.
    :return: A User instance.
    """
    return User(
        user_id=model.user_id,
        role=WorkspaceRole[str(model.role.value)],
        email=model.email,
        full_name=model.full_name,
    )


def instance_user_model(model: BaseInstanceUserResponse) -> InstanceUser:
    """
    Parse an InstanceUser from the generated model.
    :param model: The model returned by the generated code.
    :return: An InstanceUser instance.
    """
    return InstanceUser(
        user_id=model.id,
        roles=[InstanceRole(role_id=role.id, name=role.name, description=role.description) for role in model.roles],
    )


def instance_user_with_email_model(model: BaseInstanceUserResponse) -> InstanceUserWithEmail:
    """
    Parse an InstanceUserWithEmail from the generated model.
    :param model: The model returned by the generated code.
    :return: An InstanceUserWithEmail instance.
    """
    return InstanceUserWithEmail(
        email=model.email,
        full_name=model.full_name,
        user_id=model.id,
        roles=[InstanceRole(role_id=role.id, name=role.name, description=role.description) for role in model.roles],
    )


def instance_user_invitation_model(model: ListInstanceUserInvitationsResponse) -> InstanceUserInvitation:
    """
    Parse an InstanceUserInvitation from the generated model.
    :param model: The model returned by the generated code.
    :return: An InstanceUserInvitation instance.
    """
    return InstanceUserInvitation(
        email=model.email,
        invitation_id=model.id,
        roles=[InstanceRole(role_id=role.id, name=role.name, description=role.description) for role in model.roles],
        invited_at=model.created_date.replace(tzinfo=timezone.utc),
        expiration_date=model.expiration_date.replace(tzinfo=timezone.utc),
        invited_by=model.invited_by_email,
        status=model.status,
    )


def instance_user_role_model(model: ListInstanceRolesResponse) -> InstanceRoleWithPermissions:
    """
    Parse an InstanceUserRoleWithPermissions from the generated model.
    :param model: The model returned by the generated code.
    :return: An InstanceUserRoleWithPermissions instance.
    """
    return InstanceRoleWithPermissions(
        role_id=model.id, name=model.name, description=model.description, permissions=model.permissions
    )


def add_instance_user_model(model: AddInstanceUsersResponse):
    return AddedInstanceUsers(
        members=[instance_user_with_email_model(user) for user in model.members],
        invitations=[instance_user_invitation_model(invitation) for invitation in model.invitations],
    )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from collections.abc import Callable, Iterator, Sequence
from threading import Lock
from typing import TypeVar

from ..interfaces import IFeedback

__all__ = [
    "NoFeedback",
    "PartialFeedback",
    "create_default_feedback",
    "iter_with_fb",
    "reset_feedback_factory",
    "set_feedback_factory",
    "split_feedback",
]

_N_DIGITS = 4  # Let's maintain a user-friendly number of dp.


class _NoFeedback(IFeedback):
    def progress(self, progress: float, message: str | None = None) -> None:
        pass


NoFeedback = _NoFeedback()
"""A default feedback object that does nothing. Use this when no feedback is needed."""


def _no_feedback_factory(_label: str) -> IFeedback:
    return NoFeedback


_feedback_factory: Callable[[str], IFeedback] = _no_feedback_factory


def set_feedback_factory(factory: Callable[[str], IFeedback]) -> None:
    """Set the factory used by :func:`create_default_feedback`.

    This is intended to be called by the IPython extension (``%load_ext evo.widgets``)
    so that SDK operations automatically display a progress widget when running in a notebook.

    :param factory: A callable that accepts a label string and returns an :class:`IFeedback` instance.
    """
    global _feedback_factory
    _feedback_factory = factory


def reset_feedback_factory() -> None:
    """Reset the feedback factory to the default (returns :data:`NoFeedback`)."""
    global _feedback_factory
    _feedback_factory = _no_feedback_factory


def create_default_feedback(label: str) -> IFeedback:
    """Create a feedback instance using the currently registered factory.

    When no factory has been registered (or after :func:`reset_feedback_factory`),
    this returns :data:`NoFeedback`.

    :param label: A descriptive label for the operation (e.g. ``"Tasks"``, ``"Uploading data"``).
    :returns: An :class:`IFeedback` instance.
    """
    return _feedback_factory(label)


class PartialFeedback(IFeedback):
    """A wrapper for IFeedback objects that subdivides the feedback range."""

    def __init__(self, parent: IFeedback, start: float, end: float) -> None:
        """
        :param parent: The parent feedback object to subdivide.
        :param start: The start of the partial feedback range.
        :param end: The end of the partial feedback range.
        """
        self.__offset = start
        self.__factor = end - start

        self.__lock = Lock()
        self.__parent = parent

    def progress(self, progress: float, message: str | None = None) -> None:
        """Update the progress of the feedback object.

        :param progress: The progress value, between 0 and 1.
        :param message: An optional message to display.
        """
        partial_progress = round(self.__offset + (progress * self.__factor), ndigits=_N_DIGITS)
        with self.__lock:
            self.__parent.progress(partial_progress, message)


T = TypeVar("T")


def iter_with_fb(elements: Sequence[T], feedback: IFeedback | None = None) -> Iterator[tuple[T, IFeedback]]:
    """Iterate over a sequence of elements, dividing feedback uniformly throughout the range.

    :param elements: The sequence of elements to iterate over.
    :param feedback: A feedback object to subdivide.

    :yields: A tuple containing a sequence element, and a new feedback object for that element.
    """
    if len(elements) == 0:
        return

    fb_part_size = 1 / len(elements)
    for i, element in enumerate(elements):
        if feedback is None:
            yield element, NoFeedback
        else:
            start = round(i * fb_part_size, ndigits=_N_DIGITS)
            end = round((i + 1) * fb_part_size, ndigits=_N_DIGITS)
            yield element, PartialFeedback(feedback, start, end)
            feedback.progress(end)


class _ConcurrentFeedback(IFeedback):
    """A feedback object that reports progress to a ConcurrentFeedbackGroup, allowing aggregation of progress from
    multiple sources concurrently.
    """

    def __init__(self, callback: Callable[[float, str | None], None]) -> None:
        self.__callback = callback
        self.__progress = 0.0

    def progress(self, progress: float, message: str | None = None) -> None:
        """Update the progress of the feedback object.

        :param progress: The progress value, between 0 and 1.
        :param message: An optional message to display.
        """
        diff = progress - self.__progress
        self.__progress = progress
        self.__callback(diff, message)


class _ConcurrentFeedbackGroup:
    """A group of feedbacks that can be updated concurrently, aggregating their progress."""

    def __init__(self, parent: IFeedback) -> None:
        self.__lock = Lock()
        self.__parent = parent
        self.__total_weight = 0.0
        self.__progress = 0.0
        self.__n_feedbacks = 0

    def _update(self, message: str | None = None) -> None:
        if self.__total_weight == 0:
            progress = self.__progress / self.__n_feedbacks
        else:
            progress = self.__progress / self.__total_weight
        self.__parent.progress(round(progress, _N_DIGITS), message)

    def _progress(self, weight: float, progress_change: float, message: str | None) -> None:
        with self.__lock:
            if self.__total_weight == 0:
                self.__progress += progress_change
            else:
                self.__progress += weight * progress_change
            self._update(message)

    def create_feedback(self, weight: float) -> IFeedback:
        with self.__lock:
            if self.__progress > 0:
                raise RuntimeError("Cannot create new feedback after progress has been reported.")
            self.__total_weight += weight
            self.__n_feedbacks += 1

        return _ConcurrentFeedback(functools.partial(self._progress, weight))


def split_feedback(feedback: IFeedback, weights: Sequence[float]) -> list[IFeedback]:
    """Split a feedback object into multiple feedback objects based on weights.

    Each of them can be updated concurrently, and their progress will be aggregated into the parent feedback object.

    Note, if all weights are zero, all feedback objects will be uniformly weighted.

    :param feedback: The parent feedback object to subdivide.
    :param weights: Weights for each partial feedback object.
    :returns: A list of ConcurrentFeedback objects.
    """
    group = _ConcurrentFeedbackGroup(feedback)
    return [group.create_feedback(weight) for weight in weights]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .connector import APIConnector, NoAuth
from .context import StaticContext
from .data import (
    DependencyStatus,
    EmptyResponse,
    Environment,
    HealthCheckType,
    HTTPHeaderDict,
    HTTPResponse,
    Page,
    RequestMethod,
    ResourceMetadata,
    ServiceHealth,
    ServiceStatus,
    ServiceUser,
)
from .interfaces import IAuthorizer, ICache, IContext, IFeedback, ITransport
from .service import BaseAPIClient
from .typed import BoundingBox, Point3, Size3d, Size3i

__all__ = [
    "APIConnector",
    "BaseAPIClient",
    "BoundingBox",
    "DependencyStatus",
    "EmptyResponse",
    "Environment",
    "EvoContext",
    "HTTPHeaderDict",
    "HTTPResponse",
    "HealthCheckType",
    "IAuthorizer",
    "ICache",
    "IContext",
    "IFeedback",
    "ITransport",
    "NoAuth",
    "Page",
    "Point3",
    "RequestMethod",
    "ResourceMetadata",
    "ServiceHealth",
    "ServiceStatus",
    "ServiceUser",
    "Size3d",
    "Size3i",
    "StaticContext",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Workspaces API
=============

The Workspaces API enables users to organize, maintain, and store project data, but does not use or process the data. The workspace APIs allow you to manage:
- Workspaces
- User roles within workspaces
- Workspace thumbnails

There are three pre-defined roles within workspaces:

- Owner: can perform all actions in the workspace
- Editor: can perform all actions excluding deleting of a workspace
- Viewer: can view the workspace

These user roles can be assigned to users in a workspace. Once a role has been assigned it can be replaced or removed.
Users can also retrieve user roles, the role of a particular user, and their own role if applicable.
For more information on using the Workspaces API, see the [Workspaces API overview](https://developer.seequent.com/docs/guides/workspaces/), or the API references here.


This code is generated from the OpenAPI specification for Workspaces API.
API version: 1.0
"""

from evo.common.connector import APIConnector
from evo.common.data import RequestMethod
from evo.common.utils import get_header_metadata

from ..models import *  # noqa: F403

__all__ = ["LicenseAccessApi"]


class LicenseAccessApi:
    """API client for the LicenseAccess endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def v2_license_access_evo_identity_v2_license_access_get(
        self,
        service: list[str] | None = None,
        hub: str | None = None,
        org_id: str | None = None,
        required_scope: list[str] | None = None,
        tx_uuid: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> LicenseAccessResponseModel:  # noqa: F405
        """V2 License Access

        Verify the user and token has access to the requested resources.  Given a user token, verify that the user has access to the requested service, org and hub, as well as that the token as the required scope.  On success, returns details of the calling user, what they are entitled to, and the authorization the request was made with.  Args:     org_id (UUID): UUID of the Evo organisation     service (str): service name (i.e. \"blockmodel\")     hub (str): the hub code     required_scope(str): scope required for the service Returns:     A JSON response.

        :param service: (optional)
            Example: `['service_example']`
        :param hub: (optional)
            Example: `'hub_example'`
        :param org_id: (optional)
            Example: `'org_id_example'`
        :param required_scope: (optional)
            Example: `['required_scope_example']`
        :param tx_uuid: (optional)
            Example: `'tx_uuid_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the query parameters.
        _query_params = {}
        if service is not None:
            _query_params["service"] = service
        if hub is not None:
            _query_params["hub"] = hub
        if org_id is not None:
            _query_params["org_id"] = org_id
        if required_scope is not None:
            _query_params["required_scope"] = required_scope

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if tx_uuid is not None:
            _header_params["tx-uuid"] = tx_uuid
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {
            "service": "multi",
            "required_scope": "multi",
        }

        _response_types_map = {
            "200": LicenseAccessResponseModel,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/evo/identity/v2/license-access",
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from importlib.resources import files

try:
    # Traversable was moved from importlib.abc to importlib.resources.abc in Python 3.11, and
    # the re-export was removed in Python 3.14.
    from importlib.resources.abc import Traversable
except ImportError:
    from importlib.abc import Traversable


_ROOT = files(__name__)


def get(filename: str) -> Traversable:
    """Get the path to a file in this directory.

    :param filename: The name of the file.

    :return: A Traversable object representing the file.
    """
    return _ROOT / filename



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
from typing import TypeAlias

from evo import logging

from .exceptions import FileNameTooLongError

__all__ = [
    "PathLike",
    "resolve_path",
]

logger = logging.getLogger("types")

PathLike: TypeAlias = str | os.PathLike | Path


MAX_WINDOWS_PATH_LEN = 259
FILE_NAME_TOO_LONG_ERR_NO = 63


def _check_path_length(path: Path) -> None:
    if os.name == "posix":
        try:
            # https://linux.die.net/man/3/pathconf
            max_path_length = os.pathconf(path, "PC_PATH_MAX")
        except FileNotFoundError:
            # If the path does not exist, we cannot check its length
            logger.info("File does not exist.", exc_info=True)
            max_path_length = 1024  # Default value for POSIX systems
        except ValueError:
            logger.error("Error checking path length. Assuming path is not too long.", exc_info=True)
            return
        except OSError as e:
            if e.errno == FILE_NAME_TOO_LONG_ERR_NO:
                raise FileNameTooLongError() from e

            logger.error("Error checking path length. Assuming path is not too long.", exc_info=True)
            return

    elif os.name == "nt":
        max_path_length = MAX_WINDOWS_PATH_LEN
    else:
        logger.error("Unsupported operating system. Assuming path is not too long.")
        return

    if len(str(path)) > max_path_length:
        raise FileNameTooLongError()


def resolve_path(path: PathLike, *, check_path_length: bool = False) -> Path:
    """Resolve the given path and check for maximum path length.

    :param path: The path to resolve.
    :param check_path_length: If True, check the path length.

    :returns: The resolved path.

    :raises TypeError: If the path is not a PathLike object.
    :raises FileNameTooLongError: If check_path_length is True and the path is too long.
    """
    if not isinstance(path, Path):
        try:
            path = Path(path)
        except TypeError:
            raise TypeError(f"Expected a PathLike object, got {path!r}")

    resolved = path.resolve()
    if check_path_length:
        _check_path_length(resolved)
    return resolved



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# generated by datamodel-codegen:
#   filename:  spec.yaml

from __future__ import annotations

from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Annotated, Any
from uuid import UUID

from pydantic import (
    AnyUrl,
    Field,
    RootModel,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)

from .._model_config import CustomBaseModel


class BaseInstanceUserRoleResponse(CustomBaseModel):
    description: Annotated[StrictStr, Field(title="Description")]
    id: Annotated[UUID, Field(title="Id")]
    name: Annotated[StrictStr, Field(title="Name")]


class BasicWorkspaceResponse(CustomBaseModel):
    id: Annotated[UUID, Field(title="Id")]
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the workspace, unique within an organization and hub
    """


class Coordinate(RootModel[list[StrictFloat | StrictInt]]):
    root: Annotated[list[StrictFloat | StrictInt], Field(max_length=2, min_length=2)]
    """
    Coordinate as [longitude, latitude]. First one MUST be the longitude.
    """


class CoordinateSystemEntry(CustomBaseModel):
    title: Annotated[StrictStr, Field(title="Title")]
    well_known_id: Annotated[StrictStr, Field(title="Well Known Id")]


class Label(RootModel[StrictStr]):
    root: StrictStr


class ErrorInvalidParam(CustomBaseModel):
    name: Annotated[StrictStr, Field(title="Name")]
    reason: Annotated[StrictStr, Field(title="Reason")]


class ErrorResponse(CustomBaseModel):
    detail: Annotated[StrictStr | None, Field(title="Detail")] = None
    invalid_params: Annotated[
        list[ErrorInvalidParam] | None,
        Field(alias="invalid-params", title="Invalid-Params"),
    ] = None
    status: Annotated[StrictInt, Field(title="Status")]
    title: Annotated[StrictStr, Field(title="Title")]
    type: Annotated[StrictStr, Field(title="Type")]


class FolderCreateRequest(CustomBaseModel):
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the folder
    """
    parent_folder_id: Annotated[UUID | None, Field(title="Parent Folder Id")] = None
    """
    The ID of the parent folder
    """
    parent_folder_path: Annotated[Path | None, Field(title="Parent Folder Path")] = None
    """
    The path of the parent folder
    """


class FolderMoveRequest(CustomBaseModel):
    parent_folder_id: Annotated[UUID, Field(title="Parent Folder Id")]
    """
    The ID of the new parent folder
    """


class FolderUpdateRequest(CustomBaseModel):
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The new name of the folder
    """


class GeometryTypeEnum(Enum):
    Polygon = "Polygon"


class InstanceRoleWithPermissions(CustomBaseModel):
    description: Annotated[StrictStr, Field(title="Description")]
    id: Annotated[UUID, Field(title="Id")]
    name: Annotated[StrictStr, Field(title="Name")]
    permissions: Annotated[list[StrictStr], Field(title="Permissions")]


class LicenseAccessAuthDetails(CustomBaseModel):
    """
    Auth details,retrieved from token, for v2 License Access Response
    """

    client_id: Annotated[StrictStr, Field(title="Client Id")]
    exp: Annotated[StrictInt, Field(title="Exp")]
    iat: Annotated[StrictInt, Field(title="Iat")]
    iss: Annotated[StrictStr | None, Field(title="Iss")] = None
    jti: Annotated[StrictStr | None, Field(title="Jti")] = None
    scope: Annotated[list[StrictStr] | None, Field(title="Scope")] = None
    sid: Annotated[StrictStr | None, Field(title="Sid")] = None
    type: Annotated[StrictStr, Field(title="Type")]
    ver: Annotated[StrictStr, Field(title="Ver")]


class LicenseAccessEntitlements(CustomBaseModel):
    """
    Entitlement details for v2 License Access Response
    """

    hub_code: Annotated[StrictStr, Field(title="Hub Code")]
    instance_id: Annotated[StrictStr | None, Field(title="Instance Id")] = None
    instance_name: Annotated[StrictStr | None, Field(title="Instance Name")] = None
    org_id: Annotated[StrictStr, Field(title="Org Id")]
    org_name: Annotated[StrictStr, Field(title="Org Name")]
    roles: Annotated[list[StrictStr], Field(title="Roles")]
    services: Annotated[list[StrictStr], Field(title="Services")]


class LicenseAccessUser(CustomBaseModel):
    """
    User details for v2 License Access Response
    """

    org_id: Annotated[StrictStr | None, Field(title="Org Id")] = None
    org_name: Annotated[StrictStr | None, Field(title="Org Name")] = None
    sub: Annotated[StrictStr, Field(title="Sub")]


class ListInstanceRolesResponse(CustomBaseModel):
    roles: Annotated[list[InstanceRoleWithPermissions], Field(title="Roles")]


class Organization(CustomBaseModel):
    display_name: Annotated[StrictStr, Field(title="Display Name")]
    id: Annotated[StrictStr, Field(title="Id")]


class OrganizationSettingsFieldResponse(CustomBaseModel):
    ml_enabled: Annotated[StrictBool, Field(title="Ml Enabled")] = False


class OrganizationSettingsResponse(CustomBaseModel):
    created_at: Annotated[datetime | None, Field(title="Created At")] = None
    created_by: Annotated[UUID | None, Field(title="Created By")] = None
    id: Annotated[UUID, Field(title="Id")]
    settings: Annotated[OrganizationSettingsFieldResponse, Field()] = {"ml_enabled": False}
    updated_at: Annotated[datetime | None, Field(title="Updated At")] = None
    updated_by: Annotated[UUID | None, Field(title="Updated By")] = None


class PaginationLinks(CustomBaseModel):
    count: Annotated[StrictInt, Field(title="Count")]
    first: Annotated[AnyUrl, Field(title="First")]
    last: Annotated[AnyUrl, Field(title="Last")]
    next: Annotated[AnyUrl | None, Field(title="Next")] = None
    previous: Annotated[AnyUrl | None, Field(title="Previous")] = None
    total: Annotated[StrictInt, Field(title="Total")]


class PaginationLinksWithoutTotal(CustomBaseModel):
    count: Annotated[StrictInt, Field(title="Count")]
    first: Annotated[AnyUrl, Field(title="First")]
    next: Annotated[AnyUrl | None, Field(title="Next")] = None
    previous: Annotated[AnyUrl | None, Field(title="Previous")] = None


class RoleEnum(Enum):
    owner = "owner"
    editor = "editor"
    viewer = "viewer"


class Service(CustomBaseModel):
    code: Annotated[StrictStr, Field(title="Code")]
    display_name: Annotated[StrictStr, Field(title="Display Name")]


class ServiceAccess(CustomBaseModel):
    hub_code: Annotated[StrictStr, Field(title="Hub Code")]
    org_id: Annotated[StrictStr, Field(title="Org Id")]
    services: Annotated[list[StrictStr], Field(title="Services")]


class UpdateInstanceUserRolesRequest(CustomBaseModel):
    roles: Annotated[list[UUID], Field(title="Roles")]


class UpdateInstanceUserRolesResponse(CustomBaseModel):
    id: Annotated[UUID, Field(title="Id")]
    roles: Annotated[list[BaseInstanceUserRoleResponse], Field(title="Roles")]


class User(CustomBaseModel):
    email: Annotated[StrictStr | None, Field(title="Email")] = None
    full_name: Annotated[StrictStr | None, Field(title="Full Name")] = None
    role: RoleEnum
    user_id: Annotated[UUID, Field(title="User Id")]


class UserModel(CustomBaseModel):
    email: Annotated[StrictStr | None, Field(title="Email")] = None
    id: Annotated[UUID, Field(title="Id")]
    name: Annotated[StrictStr | None, Field(title="Name")] = None


class UserRole(CustomBaseModel):
    role: RoleEnum
    user_id: Annotated[UUID, Field(title="User Id")]


class UserRoleAssignmentRequest(CustomBaseModel):
    role: RoleEnum
    user_id: Annotated[UUID, Field(title="User Id")]
    workspace_id: Annotated[UUID, Field(title="Workspace Id")]


class UserRoleMapping(CustomBaseModel):
    email: Annotated[StrictStr, Field(title="Email")]
    roles: Annotated[list[UUID], Field(title="Roles")]


class UserRoleViaEmail(CustomBaseModel):
    email: Annotated[StrictStr, Field(title="Email")]
    role: RoleEnum


class WorkspaceMlEnablementRequest(CustomBaseModel):
    ml_enabled: Annotated[StrictBool, Field(title="Ml Enabled")]
    workspace_id: Annotated[UUID, Field(title="Workspace Id")]


class AddInstanceUsersRequest(CustomBaseModel):
    users: Annotated[list[UserRoleMapping], Field(title="Users")]


class AssignRoleRequest(RootModel[UserRole | UserRoleViaEmail]):
    root: Annotated[UserRole | UserRoleViaEmail, Field(title="AssignRoleRequest")]


class BaseInstanceUserInvitationResponse(CustomBaseModel):
    created_date: Annotated[datetime, Field(title="Created Date")]
    email: Annotated[StrictStr, Field(title="Email")]
    expiration_date: Annotated[datetime, Field(title="Expiration Date")]
    id: Annotated[UUID, Field(title="Id")]
    invited_by_email: Annotated[StrictStr, Field(title="Invited By Email")]
    roles: Annotated[list[BaseInstanceUserRoleResponse], Field(title="Roles")]
    status: Annotated[StrictStr, Field(title="Status")]


class BaseInstanceUserResponse(CustomBaseModel):
    email: Annotated[StrictStr, Field(title="Email")]
    full_name: Annotated[StrictStr, Field(title="Full Name")]
    id: Annotated[UUID, Field(title="Id")]
    roles: Annotated[list[BaseInstanceUserRoleResponse], Field(title="Roles")]


class BoundingBox(CustomBaseModel):
    coordinates: Annotated[list[list[Coordinate]], Field(title="Coordinates")]
    type: GeometryTypeEnum


class BulkUserRoleAssignmentsRequest(CustomBaseModel):
    role_assignments: Annotated[list[UserRoleAssignmentRequest], Field(max_length=100, title="Role Assignments")]


class CoordinateSystemCategory(CustomBaseModel):
    items: Annotated[list[CoordinateSystemCategory | CoordinateSystemEntry], Field(title="Items")]
    title: Annotated[StrictStr, Field(title="Title")]


class CreateWorkspaceRequest(CustomBaseModel):
    bounding_box: BoundingBox | None = None
    default_coordinate_system: Annotated[StrictStr, Field(title="Default Coordinate System")] = ""
    description: Annotated[StrictStr, Field(title="Description")] = ""
    labels: Annotated[list[Label] | None, Field(max_length=20, title="Labels")] = None
    """
    A list of labels
    """
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the workspace, unique within an organization and hub
    """


class Hub(CustomBaseModel):
    code: Annotated[StrictStr, Field(title="Code")]
    display_name: Annotated[StrictStr, Field(title="Display Name")]
    url: Annotated[StrictStr, Field(title="Url")]


class DiscoveryResponseContent(CustomBaseModel):
    hubs: Annotated[list[Hub], Field(title="Hubs")]
    organizations: Annotated[list[Organization], Field(title="Organizations")]
    service_access: Annotated[list[ServiceAccess], Field(title="Service Access")]
    services: Annotated[list[Service], Field(title="Services")]


class FolderResponse(CustomBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: UserModel
    deleted: Annotated[StrictBool, Field(title="Deleted")]
    folder_id: Annotated[UUID, Field(title="Folder Id")]
    name: Annotated[StrictStr, Field(title="Name")]
    parent_folder_id: Annotated[UUID | None, Field(title="Parent Folder Id")] = None
    path: Annotated[Path, Field(title="Path")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    updated_by: UserModel
    workspace_id: Annotated[UUID, Field(title="Workspace Id")]


class LicenseAccessResponseModel(CustomBaseModel):
    """
    v2 License Access Response
    """

    auth_details: LicenseAccessAuthDetails
    entitlements: LicenseAccessEntitlements
    user: LicenseAccessUser


class ListCoordinateSystemsResponse(CustomBaseModel):
    links: Annotated[dict[str, Any], Field(title="Links")]
    results: Annotated[list[CoordinateSystemCategory], Field(title="Results")]


class ListInstanceUserInvitationsResponse(CustomBaseModel):
    links: PaginationLinksWithoutTotal
    results: Annotated[list[BaseInstanceUserInvitationResponse], Field(title="Results")]


class ListInstanceUsersResponse(CustomBaseModel):
    links: PaginationLinksWithoutTotal
    results: Annotated[list[BaseInstanceUserResponse], Field(title="Results")]


class ListUserRoleResponse(CustomBaseModel):
    links: Annotated[dict[str, Any], Field(title="Links")]
    results: Annotated[list[User], Field(title="Results")]


class ListWorkspaceSummaryResponse(CustomBaseModel):
    links: PaginationLinks
    results: Annotated[list[BasicWorkspaceResponse], Field(title="Results")]


class MlEnablementRequest(CustomBaseModel):
    ml_enablements: Annotated[
        list[WorkspaceMlEnablementRequest],
        Field(max_length=100, min_length=1, title="Ml Enablements"),
    ]


class UpdateWorkspaceRequest(CustomBaseModel):
    bounding_box: BoundingBox | None = None
    default_coordinate_system: Annotated[StrictStr | None, Field(title="Default Coordinate System")] = None
    description: Annotated[StrictStr | None, Field(title="Description")] = None
    labels: Annotated[list[Label] | None, Field(max_length=20, title="Labels")] = None
    """
    A list of labels
    """
    name: Annotated[StrictStr | None, Field(max_length=60, min_length=1, title="Name")] = None


class UserWorkspaceResponse(CustomBaseModel):
    bounding_box: BoundingBox | None = None
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: UserModel
    default_coordinate_system: Annotated[StrictStr, Field(title="Default Coordinate System")] = ""
    description: Annotated[StrictStr, Field(title="Description")] = ""
    id: Annotated[UUID, Field(title="Id")]
    labels: Annotated[list[StrictStr], Field(title="Labels")] = []
    ml_enabled: Annotated[StrictBool, Field(title="Ml Enabled")] = False
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the workspace, unique within an organization and hub
    """
    self_link: Annotated[AnyUrl, Field(title="Self Link")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    updated_by: UserModel
    user_role: RoleEnum


class WorkspaceRoleOptionalResponse(CustomBaseModel):
    bounding_box: BoundingBox | None = None
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: UserModel
    current_user_role: RoleEnum | None = None
    default_coordinate_system: Annotated[StrictStr, Field(title="Default Coordinate System")] = ""
    description: Annotated[StrictStr, Field(title="Description")] = ""
    id: Annotated[UUID, Field(title="Id")]
    labels: Annotated[list[StrictStr], Field(title="Labels")] = []
    ml_enabled: Annotated[StrictBool, Field(title="Ml Enabled")] = False
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the workspace, unique within an organization and hub
    """
    self_link: Annotated[AnyUrl, Field(title="Self Link")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    updated_by: UserModel


class WorkspaceRoleRequiredResponse(CustomBaseModel):
    bounding_box: BoundingBox | None = None
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: UserModel
    current_user_role: RoleEnum
    default_coordinate_system: Annotated[StrictStr, Field(title="Default Coordinate System")] = ""
    description: Annotated[StrictStr, Field(title="Description")] = ""
    id: Annotated[UUID, Field(title="Id")]
    labels: Annotated[list[StrictStr], Field(title="Labels")] = []
    ml_enabled: Annotated[StrictBool, Field(title="Ml Enabled")] = False
    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the workspace, unique within an organization and hub
    """
    self_link: Annotated[AnyUrl, Field(title="Self Link")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    updated_by: UserModel


class AddInstanceUsersResponse(CustomBaseModel):
    invitations: Annotated[list[BaseInstanceUserInvitationResponse], Field(title="Invitations")]
    members: Annotated[list[BaseInstanceUserResponse], Field(title="Members")]


class ChildFolderListResponse(CustomBaseModel):
    count: Annotated[StrictInt, Field(title="Count")]
    folders: Annotated[list[FolderResponse], Field(title="Folders")]
    limit: Annotated[StrictInt, Field(title="Limit")]
    offset: Annotated[StrictInt, Field(title="Offset")]
    total: Annotated[StrictInt, Field(title="Total")]


class CreatePathFoldersResponse(CustomBaseModel):
    folders: Annotated[list[FolderResponse], Field(title="Folders")]


class DiscoveryResponse(CustomBaseModel):
    discovery: DiscoveryResponseContent


class GetFolderResponse(CustomBaseModel):
    child_folders: ChildFolderListResponse
    folder: FolderResponse


class ListUserWorkspacesResponse(CustomBaseModel):
    links: PaginationLinks
    results: Annotated[list[UserWorkspaceResponse], Field(title="Results")]


class ListWorkspacesResponse(CustomBaseModel):
    links: PaginationLinks
    results: Annotated[
        list[WorkspaceRoleRequiredResponse] | list[WorkspaceRoleOptionalResponse],
        Field(title="Results"),
    ]


CoordinateSystemCategory.model_rebuild()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
import uuid
from logging import getLogger
from typing import Any, Mapping, overload

from .connector import APIConnector
from .data import Environment
from .exceptions import ContextError
from .interfaces import IAuthorizer, ICache, IContext, ITransport

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

__all__ = ["StaticContext"]

logger = getLogger(__name__)


class StaticContext(IContext):
    """Static context for Evo SDKs.

    This class includes the following information:
    - Transport and authorizer, required to connect to Evo APIs.
    - Optional cache, which will be used for caching downloads/uploads of data, and other temporary data.
    - Hub, organization, and workspace values, which will be used to determine under which hub/organization/workspace
    operations are performed within. Which of these values are required depends on the specific SDK that is used.
    """

    @overload
    def __init__(
        self,
        *,
        transport: ITransport,
        authorizer: IAuthorizer,
        additional_headers: Mapping[str, Any] | None = None,
        cache: ICache | None = None,
        hub_url: str | None = None,
        org_id: uuid.UUID | None = None,
        workspace_id: uuid.UUID | None = None,
    ):
        """Initializes an EvoContext with transport and authorizer.

        :param transport: The transport to use for API calls.
        :param authorizer: The authorizer to use for API calls.
        :param additional_headers: Additional headers to include in API calls.
        :param cache: Optional cache to use for caching data.
        :param hub_url: Optional hub URL to use for API calls.
        :param org_id: Optional organization ID to use for API calls.
        :param workspace_id: Optional workspace ID to use for API calls.
        """

    @overload
    def __init__(
        self,
        *,
        connector: APIConnector,
        cache: ICache | None = None,
        org_id: uuid.UUID | None = None,
        workspace_id: uuid.UUID | None = None,
    ):
        """Initializes an EvoContext with an APIConnector.

        :param connector: The APIConnector to use for API calls.
        :param cache: Optional cache to use for caching data.
        :param org_id: Optional organization ID to use for API calls.
        :param workspace_id: Optional workspace ID to use for API calls.
        """

    def __init__(
        self,
        *,
        transport: ITransport | None = None,
        authorizer: IAuthorizer | None = None,
        connector: APIConnector | None = None,
        additional_headers: Mapping[str, Any] | None = None,
        cache: ICache | None = None,
        hub_url: str | None = None,
        org_id: uuid.UUID | None = None,
        workspace_id: uuid.UUID | None = None,
    ):
        if connector is not None:
            if transport is not None or authorizer is not None or additional_headers is not None or hub_url is not None:
                raise ValueError(
                    "If 'connector' is provided, 'transport', 'authorizer', 'additional_headers', and 'hub_url' must not be provided."
                )
            self._connector = connector
            hub_url = connector.base_url
            self._transport = None
            self._authorizer = None
            self._additional_headers = None
        else:
            if transport is None or authorizer is None:
                raise ValueError("'transport' and 'authorizer' must be provided if 'connector' is not provided.")
            self._connector = None
            self._transport = transport
            self._authorizer = authorizer
            self._additional_headers = additional_headers
        self._cache = cache
        self._hub_url = hub_url
        self._org_id = org_id
        self._workspace_id = workspace_id

    @classmethod
    def create_copy(cls, context: IContext) -> Self:
        """Creates a static copy of the given context.

        This requires that an APIConnector can be retrieved from the provided context.

        :param context: The context to copy.
        :return: A new StaticContext with the specified cache.
        :raises ContextError: If the connector cannot be retrieved from the provided context.
        """
        try:
            environment = context.get_environment()
        except ContextError:
            try:
                org_id = context.get_org_id()
            except ContextError:
                org_id = None
            workspace_id = None
        else:
            org_id = environment.org_id
            workspace_id = environment.workspace_id
        return StaticContext(
            connector=context.get_connector(),
            cache=context.get_cache(),
            org_id=org_id,
            workspace_id=workspace_id,
        )

    def get_connector(self) -> APIConnector:
        """Gets the APIConnector for this context."""
        if self._connector is not None:
            return self._connector
        hub_url = self._hub_url
        if hub_url is None:
            raise ContextError("Can't determine hub URL for connector. Context must have the hub URL set.")
        return APIConnector(
            base_url=hub_url,
            transport=self._transport,
            authorizer=self._authorizer,
            additional_headers=self._additional_headers,
        )

    def get_cache(self) -> ICache | None:
        """Gets the cache of this context, if any."""
        return self._cache

    def get_environment(self) -> Environment:
        """Gets the Environment of this context.

        :return: The Environment.
        :raises ContextError: If the context does not have sufficient information to create an Environment.
        """
        if self._hub_url is None:
            raise ContextError("Can't determine hub URL for the environment. Context must have a hub URL set.")
        if self._org_id is None:
            raise ContextError(
                "Can't determine organization for environment. Context must have an organization ID set."
            )
        if self._workspace_id is None:
            raise ContextError("Can't determine workspace for environment. Context must have a workspace ID set.")
        return Environment(
            hub_url=self._hub_url,
            org_id=self._org_id,
            workspace_id=self._workspace_id,
        )

    def get_org_id(self) -> uuid.UUID:
        """Gets the organization ID associated with this context.

        :return: The organization ID.
        :raises ContextError: If the context does not have an organization ID.
        """
        if self._org_id is None:
            raise ContextError("Can't determine organization ID. Context must have an organization ID set.")
        return self._org_id

    @classmethod
    def from_environment(
        cls,
        environment: Environment,
        connector: APIConnector,
        cache: ICache | None = None,
    ) -> Self:
        """Constructs an EvoContext from an Environment and an APIConnector."""
        if connector.base_url != environment.hub_url.rstrip("/") + "/":
            raise ContextError("The connector's base URL does not match the environment's hub URL.")
        return cls(
            connector=connector,
            cache=cache,
            org_id=environment.org_id,
            workspace_id=environment.workspace_id,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any
from urllib.parse import urlencode

import aiohttp
from aiohttp.typedefs import StrOrURL

from evo.common import HTTPHeaderDict, HTTPResponse
from evo.common.exceptions import RetryError, TransportError
from evo.common.utils import Retry

__all__ = ["Config", "Context"]


@dataclass(frozen=True, kw_only=True)
class Config:
    user_agent: str
    """The value to provide in the `User-Agent` header."""

    num_pools: int
    """Number of connection pools to cache before discarding the least recently used pool."""

    verify_ssl: bool
    """Verify SSL certificates."""

    retry: Retry
    """Retry handler."""

    proxy: StrOrURL | None
    """Proxy server to use for the request."""

    close_grace_period_ms: int
    """Grace period (in milliseconds) to wait for connections to close gracefully. 250 is used if not set"""

    async def create_context(self) -> Context:
        return Context(
            num_pools=self.num_pools,
            verify_ssl=self.verify_ssl,
            retry_handler=self.retry,
            proxy=self.proxy,
            close_grace_period_ms=self.close_grace_period_ms,
        )


class Context:
    """Inner class to manage the aiohttp session."""

    def __init__(
        self, num_pools: int, verify_ssl: bool, retry_handler: Retry, proxy: StrOrURL | None, close_grace_period_ms: int
    ) -> None:
        """
        :param num_pools: Number of connection pools to cache before discarding the least recently used pool.
        :param verify_ssl: Verify SSL certificates.
        :param retry_handler: Retry handler.
        """
        connector = aiohttp.TCPConnector(ssl=None if verify_ssl else False, limit=num_pools)

        # TODO: Debug tracing? https://docs.aiohttp.org/en/stable/tracing_reference.html
        self.__session: aiohttp.ClientSession | None = aiohttp.ClientSession(
            connector=connector, skip_auto_headers=["Accept", "Accept-Encoding"]
        )
        self.__retry_handler = retry_handler
        self.__proxy = proxy
        self._close_grace_period_ms = close_grace_period_ms

    async def close(self) -> None:
        """Close the aiohttp session."""

        session = self.__session
        self.__session = None
        await session.close()

        # Wait for the underlying SSL connections to close
        # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
        # Per comments on https://github.com/aio-libs/aiohttp/issues/1925, this should be fixed in aiohttp 4.0.
        await asyncio.sleep(self._close_grace_period_ms / 1000)

    async def __perform_request(self, **kwargs: Any) -> HTTPResponse:
        """Perform an HTTP request."""

        if self.__session is None:
            raise TransportError("Cannot make a request after the transport has been closed.")

        try:
            async for request_attempt in self.__retry_handler:
                with request_attempt.suppress_errors():
                    async with self.__session.request(allow_redirects=False, **kwargs) as resp:
                        return HTTPResponse(
                            status=resp.status,
                            data=await resp.read(),
                            reason=resp.reason,
                            headers=HTTPHeaderDict(resp.headers),
                        )
        except RetryError as error:
            raise TransportError("Reached maximum number of retries", caused_by=error).with_traceback(
                error.__traceback__
            )

    async def request_form(
        self,
        method: str,
        url: str,
        headers: HTTPHeaderDict,
        fields: list[tuple[str, str | bytes]],
        timeout: aiohttp.ClientTimeout | None,
    ) -> HTTPResponse:
        """Submit a request with urlencoded or multipart form data.

        :param method: HTTP method.
        :param url: Request URL.
        :param headers: Request headers.
        :param fields: Request form data.
        :param timeout: Request timeout.

        :return: The server response.
        """
        match headers.get("Content-Type"):
            case "multipart/form-data":
                data = aiohttp.FormData(fields, quote_fields=False)
            case "application/x-www-form-urlencoded":
                data = urlencode(fields)
            case content_type:
                raise NotImplementedError(f"Unsupported form content type '{content_type}'")

        return await self.__perform_request(
            method=method, url=url, data=data, headers=headers, timeout=timeout, proxy=self.__proxy
        )

    async def request(
        self,
        method: str,
        url: str,
        headers: HTTPHeaderDict,
        body: str | bytes | None,
        timeout: aiohttp.ClientTimeout | None,
    ) -> HTTPResponse:
        """Submit a standard HTTP request.

        :param method: HTTP method.
        :param url: Request URL.
        :param headers: Request headers.
        :param body: Serialized request body.
        :param timeout: Request timeout.

        :return: The server response.
        """
        return await self.__perform_request(
            method=method, url=url, headers=headers, data=body, timeout=timeout, proxy=self.__proxy
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections.abc import Sequence
from uuid import UUID

from pydantic import BaseModel, field_validator

from evo.common import APIConnector, RequestMethod

from .data import Hub, Organization

__all__ = ["DiscoveryAPIClient"]


class _ServiceDiscoveryHub(BaseModel):
    url: str
    code: str
    display_name: str


class _ServiceDiscoveryOrganization(BaseModel):
    id: UUID
    display_name: str


class _ServiceDiscoveryAccess(BaseModel):
    hub_code: str
    org_id: UUID
    services: list[str]


class _ServiceDiscoveryResult(BaseModel):
    organizations: list[_ServiceDiscoveryOrganization]
    hubs: list[_ServiceDiscoveryHub]
    service_access: list[_ServiceDiscoveryAccess]

    @field_validator("organizations", "hubs")
    def _sort_by_display_name(cls, v: list) -> list:
        # Sort organizations and hubs alphanumerically by display name.
        return sorted(v, key=lambda i: i.display_name)

    def __contains__(self, item: tuple[_ServiceDiscoveryOrganization, _ServiceDiscoveryHub]) -> bool:
        assert isinstance(item, tuple) and len(item) == 2, "Expected a tuple containing organization and hub."
        org, hub = item
        assert isinstance(org, _ServiceDiscoveryOrganization), (
            "Expected organization to be an instance of _ServiceDiscoveryOrganization."
        )
        assert isinstance(hub, _ServiceDiscoveryHub), "Expected hub to be an instance of _ServiceDiscoveryHub."

        for access in self.service_access:
            if access.org_id == org.id and access.hub_code == hub.code:
                return True
        return False

    def get_service_access_services(self, org: _ServiceDiscoveryOrganization, hub: _ServiceDiscoveryHub) -> list[str]:
        assert (org, hub) in self, "Expected a service access entry for the specified organization and hub."
        return next(
            service_access
            for service_access in self.service_access
            if service_access.hub_code == hub.code and service_access.org_id == org.id
        ).services


class _DiscoveryResult(BaseModel):
    discovery: _ServiceDiscoveryResult


class DiscoveryAPIClient:
    """Simple client for interacting with the Discovery API."""

    def __init__(self, connector: APIConnector) -> None:
        """
        :param connector: The API connector to use for making requests.
        """
        self._connector = connector

    async def list_organizations(self, service_codes: Sequence[str] = ("evo",)) -> list[Organization]:
        """Get organizations with access to the specified services.

        :param service_codes: The service codes to use in the query.
        """
        result = await self._connector.call_api(
            RequestMethod.GET,
            "/evo/identity/v2/discovery",
            query_params={"service": service_codes},
            collection_formats={"service": "multi"},
            response_types_map={"200": _DiscoveryResult},
        )
        discovered = result.discovery
        return [
            Organization(
                id=org.id,
                display_name=org.display_name,
                hubs=tuple(
                    Hub(
                        url=hub.url,
                        code=hub.code,
                        display_name=hub.display_name,
                        services=tuple(discovered.get_service_access_services(org, hub)),
                    )
                    for hub in discovered.hubs
                    if (org, hub) in discovered
                ),
            )
            for org in discovered.organizations
        ]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import contextlib
import json
import warnings
from collections.abc import AsyncIterator
from typing import Generic, TypeVar

from evo import logging, oauth
from evo.oauth.authorizer import _BaseAuthorizer

from .env import DotEnv

__all__ = [
    "AuthorizationCodeAuthorizer",
]

logger = logging.getLogger("notebooks.oauth")

_TOKEN_KEY = "NotebookOAuth.token"

_T_token = TypeVar("_T_token", bound=oauth.AccessToken)


class _OAuthEnv(Generic[_T_token]):
    def __init__(self, env: DotEnv) -> None:
        self.__dotenv = env

    def get_token(self) -> _T_token | None:
        token_str = self.__dotenv.get(_TOKEN_KEY)
        if token_str is None:
            return None

        try:
            token_dict = json.loads(token_str)
            return oauth.AccessToken.model_validate(token_dict)

        except Exception:
            # The token is invalid.
            raise ValueError(f"Invalid token found in the environment file {_TOKEN_KEY}: {token_str!r}")

    def set_token(self, token: _T_token | None) -> None:
        if token is None:
            new_value = None
        else:
            new_value = token.model_dump_json(by_alias=True, exclude_unset=True)
        self.__dotenv.set(_TOKEN_KEY, new_value)


class _NotebookAuthorizerMixin(_BaseAuthorizer[_T_token]):
    pi_partial_implementation = True  # Indicate to pure-interface that this is a mixin.

    _env: _OAuthEnv[_T_token]

    async def reuse_token(self) -> bool:
        """Attempt to reuse an existing token from the environment file.

        :returns: True if a token was found and reused, False otherwise.
        """
        async with self._mutex:
            if (token := self._env.get_token()) is None:
                return False

            if token.is_expired:
                self._env.set_token(None)
                return False

            return True

    def _get_token(self) -> _T_token | None:
        return self._env.get_token()

    def _update_token(self, new_token: _T_token) -> None:
        super()._update_token(new_token)
        self._env.set_token(new_token)


class AuthorizationCodeAuthorizer(_NotebookAuthorizerMixin[oauth.AccessToken], oauth.AuthorizationCodeAuthorizer):
    """An authorization code authorizer for use in Jupyter notebooks.

    This authorizer is not secure, and should only ever be used in Jupyter notebooks. It stores the access token in the
    environment file, which is not secure. It is intended for use in a development environment only. The environment
    file must not be committed to source control.
    """

    def __init__(
        self,
        oauth_connector: oauth.OAuthConnector,
        redirect_url: str,
        scopes: oauth.AnyScopes,
        env: DotEnv,
    ) -> None:
        """
        :param oauth_connector: The OAuth connector to use for authentication.
        :param redirect_url: The local URL to redirect the user back to after authorisation.
        :param scopes: The OAuth scopes to request.
        :param env: The environment to store the OAuth token in.
        """
        warnings.warn(
            "The evo.notebooks.AuthorizationCodeAuthorizer is not secure, and should only ever be used in Jupyter"
            " notebooks in a private environment."
        )
        super().__init__(oauth_connector=oauth_connector, redirect_url=redirect_url, scopes=scopes)
        self._env = _OAuthEnv(env)

    @contextlib.asynccontextmanager
    async def _unwrap_token(self) -> AsyncIterator[oauth.AccessToken]:
        # Overrides the parent implementation so that we can automatically login at startup.
        async with self._mutex:
            if (token := self._env.get_token()) is None:
                token = await self._handle_login(timeout_seconds=60)
                self._update_token(token)
            yield token

    async def refresh_token(self) -> bool:
        succeeded = await oauth.AuthorizationCodeAuthorizer.refresh_token(self)
        if not succeeded:
            # The refresh token has expired. Clear the token from the environment.
            self._env.set_token(None)
        return succeeded



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import os
from collections.abc import Awaitable, Callable, Mapping
from datetime import datetime, timezone
from pathlib import Path
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import Any, ParamSpec, TypeVar
from urllib.parse import parse_qsl, urlparse

from evo import logging

from .._types import PathLike, resolve_path
from ..connector import APIConnector
from ..data import EmptyResponse, HTTPHeaderDict, HTTPResponse, RequestMethod
from ..exceptions import ForbiddenException, UnauthorizedException
from ..interfaces import IFeedback, ITransport
from ..utils import NoFeedback, Retry
from .bytes import BytesDestination
from .chunked_io_manager import ChunkedIOManager
from .exceptions import ChunkedIOError, RenewalError, RenewalTimeoutError
from .interfaces import ISource

logger = logging.getLogger("io.http")

__all__ = [
    "HTTPIOBase",
    "HTTPSource",
    "ResourceAuthorizationError",
]


T = TypeVar("T")
P = ParamSpec("P")

_T_Self = TypeVar("_T_Self", bound="HTTPIOBase")


class ResourceAuthorizationError(ChunkedIOError):
    """Resource authorization failed.

    This error is raised when resource operations fail with a 401 status code. This error usually indicates that
    the resource URL that was provided has expired and needs to be renewed. The error can be recovered by renewing the
    resource URL, which must in turn be supported by the issuing API.
    """

    def __init__(self, message: str, client: HTTPIOBase) -> None:
        """
        :param message: The error message.
        :param client: The client that raised the exception.
        """
        super().__init__(message)
        self.client = client

    async def recover(self) -> bool:
        """Renew the resource URL so that we can retry the operation.

        :return: True if the renewal was successful.
        """
        return await self.client.renew_url()


class HTTPIOBase:
    """Base class for HTTP-based IO operations."""

    _RENEWAL_SECONDS_THRESHOLD = 5 * 60  # Five minutes

    def __init__(
        self,
        url_callback: Callable[[], Awaitable[str]],
        transport: ITransport,
        additional_headers: dict[str, str] | None = None,
    ) -> None:
        """
        :param url_callback: An awaitable callback method that accepts no arguments and returns a new resource url.
            The callback is used to renew the resource url when the current one expires. Resource URLs are considered
            expired when resource operations fail with a 401 status code.
        :param transport: The transport to use for the API calls.
        """
        self.__callback = url_callback
        self.__transport = transport
        self.__mutex = asyncio.Lock()
        self.__connector: APIConnector | None = None
        self.__resource_path: str | None = None
        self.__query_params: list[tuple[str, Any]] = []
        self.__renewal_time: datetime | None = None
        self.__additional_headers = additional_headers

    async def __get_new_url(self) -> str:
        """Get a new resource url.

        :returns: The new resource url.
        """
        url = await self.__callback()
        self.__renewal_time = datetime.now(timezone.utc)
        return url

    def __update_url(self, url: str) -> None:
        """Update the resource url.

        :param url: The new resource url.
        """
        if not url.startswith(self.__connector.base_url):
            raise RenewalError("Renewed resource URL does not match the original resource URL.")
        uri = urlparse(url)
        self.__query_params = parse_qsl(uri.query, keep_blank_values=True)
        self.__resource_path = uri._replace(query="").geturl().removeprefix(self.__connector.base_url)

    async def __aenter__(self: _T_Self) -> _T_Self:
        async with self.__mutex:
            if self.__connector is None:
                url = await self.__get_new_url()
                if not url.startswith("https://"):
                    raise ValueError("Unsupported URL scheme")
                base_url = "https://" + urlparse(url).hostname
                self.__connector = APIConnector(
                    base_url=base_url, transport=self.__transport, additional_headers=self.__additional_headers
                )
                self.__update_url(url)
        await self.__connector.__aenter__()
        return self

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_val: Exception | None, exc_tb: TracebackType | None
    ) -> None:
        await self.__connector.__aexit__(exc_type, exc_val, exc_tb)

    async def renew_url(self) -> bool:
        """Get a new resource url for resource operations.

        The url_callback may not be thread-safe and should only be called in one thread. This function will not renew
        and will raise RenewalTimeoutError if the renewal was requested within a specified period of the previous
        renewal.

        Exceptions raised by the url_callback are not handled and will be propagated.

        :raises RenewalTimeoutError: if the renewal request was made within the threshold

        :return: True when the renewal is successful
        """
        async with self.__mutex:
            assert self.__connector is not None

            if self.__renewal_time is not None:
                delta = datetime.now(timezone.utc) - self.__renewal_time
                if delta.total_seconds() < self._RENEWAL_SECONDS_THRESHOLD:
                    raise RenewalTimeoutError(
                        f"Renewed the url {delta.total_seconds()} seconds ago, but {self._RENEWAL_SECONDS_THRESHOLD} "
                        "second period required."
                    )

            self.__update_url(await self.__get_new_url())

        return True

    async def _query_resource(
        self,
        method: RequestMethod,
        *,
        query_params: Mapping[str, Any] | None = None,
        header_params: Mapping[str, Any] | None = None,
        body: object | str | bytes | None = None,
        collection_formats: Mapping[str, str] | None = None,
        response_types_map: Mapping[str, type[T]] | None = None,
    ) -> T:
        """Query the resource with the given method and headers.

        :param method: HTTP request method.
        :param query_params: Query parameters to embed in the url.
        :param header_params: Header parameters to be placed in the request header.
        :param body: Body to send with the request.
        :param response_types_map: Mapping of response status codes to response data types. The response will
            be deserialized to the corresponding type.
        """
        all_query_params = self.__query_params.copy()
        if query_params is not None:
            all_query_params.extend(query_params.items())
        async with self.__mutex:
            assert self.__connector is not None
            query = self.__connector.call_api(
                method,
                self.__resource_path,
                query_params=all_query_params,
                header_params=header_params,
                body=body,
                collection_formats=collection_formats,
                response_types_map=response_types_map,
            )

        try:
            return await query
        except (UnauthorizedException, ForbiddenException) as e:
            raise ResourceAuthorizationError(str(e), self) from e
        except Exception as e:
            raise ChunkedIOError(str(e)) from e


class HTTPSource(HTTPIOBase, ISource):
    """HTTP source for reading data in chunks.

    This general-purpose client relies on HTTP range requests to read chunks of data from a remote resource.
    The proliferation of HTTP range request support in modern web servers makes this client suitable for downloading
    large files from a variety of sources.

    For more information on range requests, refer to https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests.
    """

    def __init__(
        self,
        url_callback: Callable[[], Awaitable[str]],
        transport: ITransport,
        additional_headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(url_callback, transport, additional_headers)
        self._size: int | None = None

    async def __aenter__(self) -> HTTPSource:
        await super().__aenter__()
        await self.get_size()
        return self

    async def get_size(self) -> int:
        """Get the size of the source data.

        :returns: The size of the source data in bytes.

        :raises ValueError: If the target resource does not support range requests.
        """
        if self._size is None:
            logger.debug("Requesting file size...")
            response = await self._query_resource(
                RequestMethod.HEAD,
                response_types_map={
                    "200": EmptyResponse,  # Use the raw response so that we can access the headers.
                },
            )
            headers = response.headers
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests#checking_if_a_server_supports_partial_requests
            if headers.get("Accept-Ranges", None) is None:
                raise ChunkedIOError("Requested resource does not support range requests.")
            self._size = int(headers["Content-Length"])  # Content-Length is the total size of the resource.
        return self._size

    async def read_chunk(self, offset: int, length: int) -> bytes:
        """Read a chunk of data from the source.

        :param offset: The offset from the start of the source data to read from.
        :param length: The length of the data to read.

        :returns: The data read from the source.
        """
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests#single_part_ranges
        req_headers = HTTPHeaderDict(range=f"bytes={offset}-{offset + length - 1}")
        logger.debug(f"Requesting chunk {req_headers['Range']}")
        response = await self._query_resource(
            RequestMethod.GET,
            header_params=req_headers,
            response_types_map={
                "206": HTTPResponse,  # Use the raw response so that we can access the headers.
            },
        )
        logger.debug(f"Received chunk {response.headers['Content-Range']}")
        return response.data

    @staticmethod
    async def download_file(
        filename: PathLike,
        url_generator: Callable[[], Awaitable[str]],
        transport: ITransport,
        max_workers: int | None = None,
        retry: Retry | None = None,
        overwrite: bool = False,
        additional_headers: dict[str, str] | None = None,
        fb: IFeedback = NoFeedback,
    ) -> None:
        """Download an HTTP resource to a file with the given filename.

        The url generator MUST generate a resource URL that can be used to access the required resource. The URL
        generator may be called again if the last URL expires (unless Retry is initialised with max_attempts == 0).

        :param filename: file to download to
        :param url_generator: An awaitable callback that accepts no arguments and returns a URL to download from
        :param transport: The transport to use for the API calls.
        :param max_workers: The maximum number of concurrent connections to use for data transfer. If None, the default
            is `min(32, os.cpu_count() + 4)`
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created. If a chunk is successfully
            transferred the attempt counter will be reset.
        :param overwrite: whether to overwrite an existing local file
        :param additional_headers: Additional headers to include in each request.
        :param fb: feedback to track the download, by tracking writes to the file

        :raises FileNameTooLongError: If the filename is too long.
        :raises ValueError: if the file already exists and overwrite is False
        :raises ChunkedIOException: if a non-recoverable exception occurred.
        :raises RetryError: if the maximum number of consecutive attempts have failed.
        """
        dst_path = resolve_path(filename, check_path_length=True)

        if dst_path.exists() and not overwrite:
            raise ValueError(f"file {dst_path} already exists, must specify overwrite")

        if retry is None:
            retry = Retry(logger=logger)
        manager = ChunkedIOManager(retry=retry, max_workers=max_workers)

        async with HTTPSource(url_generator, transport, additional_headers=additional_headers) as source:
            tmp_file = None
            try:
                with NamedTemporaryFile(delete=False, dir=dst_path.parent) as tmp_file:
                    logger.debug(f"Writing temporary file {tmp_file.name}")
                    destination = BytesDestination(tmp_file)
                    await manager.run(source, destination, fb)

                tmp_path = Path(tmp_file.name).resolve()

                logger.debug(f"Renaming {tmp_path.name} to {dst_path.name}")
                tmp_path.replace(dst_path)

            except BaseException:
                if tmp_file is not None:
                    logger.error(
                        f"Removing temporary file '{tmp_file.name}' due to an unhandled exception", exc_info=True
                    )
                    os.unlink(tmp_file.name)
                raise
